@@ -1,24 +1,24 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2013 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
  *   Free Software Foundation; either version 2 of the License, or (at your
  *   option) any later version.
- * 
+ *
  *   This program is distributed in the hope that it will be useful,
  *   but WITHOUT ANY WARRANTY; without even the implied warranty of
  *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  *   See the GNU General Public License for more details.
- * 
+ *
  *   You should have received a copy of the GNU General Public License along
  *   with this program; if not, see <http://www.gnu.org/licenses>.
- * 
+ *
  *   Linking stunnel statically or dynamically with other modules is making
  *   a combined work based on stunnel. Thus, the terms and conditions of
  *   the GNU General Public License cover the whole combination.
- * 
+ *
  *   In addition, as a special exception, the copyright holder of stunnel
  *   gives you permission to combine stunnel with free software programs or
  *   libraries that are released under the GNU LGPL and with code included
@@ -26,7 +26,7 @@
  *   modified versions of such code, with unchanged license). You may copy
  *   and distribute such a system following the terms of the GNU GPL for
  *   stunnel and the licenses of the other code concerned.
- * 
+ *
  *   Note that people who make modified versions of stunnel are not obligated
  *   to grant this special exception for their modified versions; it is their
  *   choice whether to do so. The GNU General Public License gives permission
@@ -83,7 +83,7 @@ int pty_allocate(int *ptyfd, int *ttyfd, char *namebuf) {
         ioerror("openpty");
         return -1;
     }
-    safecopy(namebuf, buf); /* possible truncation */
+    strcpy(namebuf, buf); /* possible truncation */
     return 0;
 #else /* HAVE_OPENPTY */
 #ifdef HAVE__GETPTY
@@ -98,7 +98,7 @@ int pty_allocate(int *ptyfd, int *ttyfd, char *namebuf) {
         ioerror("_getpty");
         return -1;
     }
-    safecopy(namebuf, slave);
+    strcpy(namebuf, slave);
     /* open the slave side */
     *ttyfd=open(namebuf, O_RDWR|O_NOCTTY);
     if(*ttyfd<0) {
@@ -133,7 +133,7 @@ int pty_allocate(int *ptyfd, int *ttyfd, char *namebuf) {
     pts=ptsname(ptm);
     if(pts==NULL)
         s_log(LOG_ERR, "Slave pty side name could not be obtained");
-    safecopy(namebuf, pts);
+    strcpy(namebuf, pts);
     *ptyfd=ptm;
 
     /* open the slave side */
@@ -166,7 +166,7 @@ int pty_allocate(int *ptyfd, int *ttyfd, char *namebuf) {
         s_log(LOG_ERR, "Open of /dev/ptc returns device for which ttyname fails");
         return -1;
     }
-    safecopy(namebuf, name);
+    strcpy(namebuf, name);
     *ttyfd=open(name, O_RDWR|O_NOCTTY);
     if(*ttyfd<0) {
         ioerror(name);
@@ -195,7 +195,7 @@ int pty_allocate(int *ptyfd, int *ttyfd, char *namebuf) {
         if(*ptyfd<0)
             continue;
 #ifdef HAVE_SNPRINTF
-        snprintf(namebuf, STRLEN,
+        snprintf(namebuf, 64,
 #else
         sprintf(namebuf,
 #endif
