@@ -1,24 +1,24 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2013 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
  *   Free Software Foundation; either version 2 of the License, or (at your
  *   option) any later version.
- * 
+ *
  *   This program is distributed in the hope that it will be useful,
  *   but WITHOUT ANY WARRANTY; without even the implied warranty of
  *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  *   See the GNU General Public License for more details.
- * 
+ *
  *   You should have received a copy of the GNU General Public License along
  *   with this program; if not, see <http://www.gnu.org/licenses>.
- * 
+ *
  *   Linking stunnel statically or dynamically with other modules is making
  *   a combined work based on stunnel. Thus, the terms and conditions of
  *   the GNU General Public License cover the whole combination.
- * 
+ *
  *   In addition, as a special exception, the copyright holder of stunnel
  *   gives you permission to combine stunnel with free software programs or
  *   libraries that are released under the GNU LGPL and with code included
@@ -26,7 +26,7 @@
  *   modified versions of such code, with unchanged license). You may copy
  *   and distribute such a system following the terms of the GNU GPL for
  *   stunnel and the licenses of the other code concerned.
- * 
+ *
  *   Note that people who make modified versions of stunnel are not obligated
  *   to grant this special exception for their modified versions; it is their
  *   choice whether to do so. The GNU General Public License gives permission
@@ -38,6 +38,22 @@
 #include "common.h"
 #include "prototypes.h"
 
+/* http://www.openssl.org/support/faq.html#PROG2 */
+#ifdef USE_WIN32
+#ifdef __GNUC__
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-pedantic"
+#endif /* __GNUC__ */
+#ifdef __GNUC__
+#include <../ms/applink.c>
+#else /* __GNUC__ */
+#include <openssl/applink.c>
+#endif /* __GNUC__ */
+#ifdef __GNUC__
+#pragma GCC diagnostic pop
+#endif /* __GNUC__ */
+#endif /* USE_WIN32 */
+
 /**************************************** prototypes */
 
 #ifdef __INNOTEK_LIBC__
@@ -48,123 +64,188 @@ struct sockaddr_un {
 };
 #endif
 
-static void accept_connection(SERVICE_OPTIONS *);
-static void get_limits(void); /* setup global max_clients and max_fds */
+#ifndef USE_WIN32
+static int main_unix(int, char*[]);
+#endif
+static int accept_connection(SERVICE_OPTIONS *);
 #ifdef HAVE_CHROOT
-static void change_root(void);
+static int change_root(void);
 #endif
 #if !defined(USE_WIN32) && !defined(__vms)
-static void daemonize(void);
-static void create_pid(void);
+static int daemonize(int);
+static int create_pid(void);
 static void delete_pid(void);
 #endif
-static int setup_fd(int, int, char *);
 #if !defined(USE_WIN32) && !defined(USE_OS2)
 static void signal_handler(int);
 #endif
-static void signal_pipe_init(void);
-static void signal_pipe_empty(void);
+static int signal_pipe_init(void);
+static int signal_pipe_dispatch(void);
 #ifdef USE_FORK
 static void client_status(void); /* dead children detected */
 #endif
 
 /**************************************** global variables */
 
-static int max_fds;
-static int max_clients=0;
 static int signal_pipe[2]={-1, -1};
 
+#ifndef USE_FORK
+int max_clients=0;
 volatile int num_clients=0; /* current number of clients */
+#endif
 s_poll_set *fds; /* file descriptors of listening sockets */
 
 /**************************************** startup */
 
 #ifndef USE_WIN32
 int main(int argc, char* argv[]) { /* execution begins here 8-) */
+    int retval;
+
+#ifdef M_MMAP_THRESHOLD
+    mallopt(M_MMAP_THRESHOLD, 4096);
+#endif
     str_init(); /* initialize per-thread string management */
-    main_initialize(argc>1 ? argv[1] : NULL, argc>2 ? argv[2] : NULL);
+    retval=main_unix(argc, argv);
+    unbind_ports();
+    s_poll_free(fds);
+    fds=NULL;
+    str_stats();
+    log_flush(LOG_MODE_ERROR);
+    return retval;
+}
+
+static int main_unix(int argc, char* argv[]) {
+#if !defined(__vms) && !defined(USE_OS2)
+    int fd;
+
+    fd=open("/dev/null", O_RDWR); /* open /dev/null before chroot */
+    if(fd<0)
+        fatal("Could not open /dev/null");
+#endif /* standard Unix */
+    main_initialize();
+    if(main_configure(argc>1 ? argv[1] : NULL, argc>2 ? argv[2] : NULL)) {
+        close(fd);
+        return 1;
+    }
     if(service_options.next) { /* there are service sections -> daemon mode */
 #if !defined(__vms) && !defined(USE_OS2)
-        if(!(global_options.option.foreground))
-            daemonize();
+        if(daemonize(fd)) {
+            close(fd);
+            return 1;
+        }
+        close(fd);
         /* create_pid() must be called after drop_privileges()
          * or it won't be possible to remove the file on exit */
         /* create_pid() must be called after daemonize()
          * since the final pid is not known beforehand */
-        create_pid();
+        if(create_pid())
+            return 1;
 #endif /* standard Unix */
-        num_clients=0;
+        signal(SIGCHLD, signal_handler); /* handle dead children */
+        signal(SIGHUP, signal_handler); /* configuration reload */
+        signal(SIGUSR1, signal_handler); /* log reopen */
+        signal(SIGPIPE, SIG_IGN); /* ignore broken pipe */
+        if(signal(SIGTERM, SIG_IGN)!=SIG_IGN)
+            signal(SIGTERM, signal_handler); /* fatal */
+        if(signal(SIGQUIT, SIG_IGN)!=SIG_IGN)
+            signal(SIGQUIT, signal_handler); /* fatal */
+        if(signal(SIGINT, SIG_IGN)!=SIG_IGN)
+            signal(SIGINT, signal_handler); /* fatal */
         daemon_loop();
     } else { /* inetd mode */
-        num_clients=1;
-        client(alloc_client_session(&service_options, 0, 1));
-        log_close();
+#if !defined(__vms) && !defined(USE_OS2)
+        close(fd);
+#endif /* standard Unix */
+        signal(SIGCHLD, SIG_IGN); /* ignore dead children */
+        signal(SIGPIPE, SIG_IGN); /* ignore broken pipe */
+        client_main(alloc_client_session(&service_options, 0, 1));
     }
-    return 0; /* success */
+    return 0;
 }
 #endif
 
-void main_initialize(char *arg1, char *arg2) {
-    ssl_init(); /* initialize SSL library */
-    sthreads_init(); /* initialize critical sections & SSL callbacks */
+void main_initialize() { /* one-time initialization */
+    /* basic initialization contains essential functions required for logging
+     * subsystem to function properly, thus all errors here are fatal */
+    if(ssl_init()) /* initialize SSL library */
+        fatal("SSL initialization failed");
+    if(sthreads_init()) /* initialize critical sections & SSL callbacks */
+        fatal("Threads initialization failed");
+#ifndef USE_FORK
     get_limits(); /* required by setup_fd() */
-
+#endif
     fds=s_poll_alloc();
-    signal_pipe_init();
-    s_poll_init(fds);
-    s_poll_add(fds, signal_pipe[0], 1, 0);
-    /* the most essential initialization is performed here,
-     * so gui.c can execute a thread with daemon_loop() */
-
+    if(signal_pipe_init())
+        fatal("Signal pipe initialization failed: "
+            "check your personal firewall");
     stunnel_info(LOG_NOTICE);
-    parse_commandline(arg1, arg2);
-    str_canary(); /* needs prng initialization from parse_commandline */
+}
+
+    /* configuration-dependent initialization */
+int main_configure(char *arg1, char *arg2) {
+    if(parse_commandline(arg1, arg2))
+        return 1;
+    str_canary_init(); /* needs prng initialization from parse_commandline */
 #if !defined(USE_WIN32) && !defined(__vms)
     /* syslog_open() must be called before change_root()
      * to be able to access /dev/log socket */
     syslog_open();
 #endif /* !defined(USE_WIN32) && !defined(__vms) */
     if(bind_ports())
-        die(1);
+        return 1;
 
 #ifdef HAVE_CHROOT
     /* change_root() must be called before drop_privileges()
      * since chroot() needs root privileges */
-    change_root();
+    if(change_root())
+        return 1;
 #endif /* HAVE_CHROOT */
 
 #if !defined(USE_WIN32) && !defined(__vms) && !defined(USE_OS2)
-    drop_privileges(1);
+    if(drop_privileges(1))
+        return 1;
 #endif /* standard Unix */
 
     /* log_open() must be be called after drop_privileges()
      * or logfile rotation won't be possible */
     /* log_open() must be be called before daemonize()
      * since daemonize() invalidates stderr */
-    log_open();
+    if(log_open())
+        return 1;
+    return 0;
 }
 
-/**************************************** main loop */
+/**************************************** main loop accepting connections */
 
 void daemon_loop(void) {
     SERVICE_OPTIONS *opt;
+    int temporary_lack_of_resources;
 
     while(1) {
-        if(s_poll_wait(fds, -1, -1)>=0) { /* non-critical error */
+        temporary_lack_of_resources=0;
+        if(s_poll_wait(fds, -1, -1)>=0) {
             if(s_poll_canread(fds, signal_pipe[0]))
-                signal_pipe_empty();
+                if(signal_pipe_dispatch()) /* received SIGNAL_TERMINATE */
+                    break; /* terminate daemon_loop */
             for(opt=service_options.next; opt; opt=opt->next)
-                if(s_poll_canread(fds, opt->fd))
-                    accept_connection(opt);
+                if(opt->option.accept && s_poll_canread(fds, opt->fd))
+                    if(accept_connection(opt))
+                        temporary_lack_of_resources=1;
         } else {
-            log_error(LOG_INFO, get_last_socket_error(),
+            log_error(LOG_NOTICE, get_last_socket_error(),
                 "daemon_loop: s_poll_wait");
+            temporary_lack_of_resources=1;
+        }
+        if(temporary_lack_of_resources) {
+            s_log(LOG_NOTICE,
+                "Accepting new connections suspended for 1 second");
             sleep(1); /* to avoid log trashing */
         }
     }
 }
 
-static void accept_connection(SERVICE_OPTIONS *opt) {
+    /* return 1 when a short delay is needed before another try */
+static int accept_connection(SERVICE_OPTIONS *opt) {
     SOCKADDR_UNION addr;
     char *from_address;
     int s;
@@ -176,8 +257,8 @@ static void accept_connection(SERVICE_OPTIONS *opt) {
         if(s>=0) /* success! */
             break;
         switch(get_last_socket_error()) {
-            case S_EINTR:
-                break; /* retry */
+            case S_EINTR: /* interrupted by a signal */
+                break; /* retry now */
             case S_EMFILE:
 #ifdef S_ENFILE
             case S_ENFILE:
@@ -188,35 +269,30 @@ static void accept_connection(SERVICE_OPTIONS *opt) {
 #ifdef S_ENOMEM
             case S_ENOMEM:
 #endif
-                sleep(1); /* temporarily out of resources - short delay */
+                return 1; /* temporary lack of resources */
             default:
-                sockerror("accept");
-                return; /* error */
+                return 0; /* any other error */
         }
     }
     from_address=s_ntop(&addr, addrlen);
-    s_log(LOG_DEBUG, "Service %s accepted FD=%d from %s",
+    s_log(LOG_DEBUG, "Service [%s] accepted (FD=%d) from %s",
         opt->servname, s, from_address);
     str_free(from_address);
+#ifndef USE_FORK
     if(max_clients && num_clients>=max_clients) {
         s_log(LOG_WARNING, "Connection rejected: too many clients (>=%d)",
             max_clients);
         closesocket(s);
-        return;
+        return 0;
     }
-    enter_critical_section(CRIT_CLIENTS); /* for multi-cpu machines */
-    /* increment before create_client() to prevent race condition
-     * resulting in logging "Service xxx finished (-1 left)" */
-    ++num_clients;
-    leave_critical_section(CRIT_CLIENTS);
-    if(create_client(opt->fd, s, alloc_client_session(opt, s, s), client)) {
+#endif
+    if(create_client(opt->fd, s,
+            alloc_client_session(opt, s, s), client_thread)) {
         s_log(LOG_ERR, "Connection rejected: create_client failed");
-        enter_critical_section(CRIT_CLIENTS); /* for multi-cpu machines */
-        --num_clients;
-        leave_critical_section(CRIT_CLIENTS);
         closesocket(s);
-        return;
+        return 0;
     }
+    return 0;
 }
 
 /**************************************** initialization helpers */
@@ -230,9 +306,14 @@ void unbind_ports(void) {
 
     s_poll_init(fds);
     s_poll_add(fds, signal_pipe[0], 1, 0);
-    for(opt=service_options.next; opt; opt=opt->next)
+
+    for(opt=service_options.next; opt; opt=opt->next) {
+        s_log(LOG_DEBUG, "Closing service [%s]", opt->servname);
         if(opt->option.accept && opt->fd>=0) {
             closesocket(opt->fd);
+            s_log(LOG_DEBUG, "Service [%s] closed (FD=%d)",
+                opt->servname, opt->fd);
+            opt->fd=-1;
 #ifdef HAVE_STRUCT_SOCKADDR_UN
             if(opt->local_addr.sa.sa_family==AF_UNIX) {
                 if(lstat(opt->local_addr.un.sun_path, &st))
@@ -247,10 +328,22 @@ void unbind_ports(void) {
                         opt->local_addr.un.sun_path);
             }
 #endif
-            s_log(LOG_DEBUG, "Service %s closed FD=%d",
-                opt->servname, opt->fd);
-            opt->fd=-1;
+        } else if(opt->option.program && opt->option.remote) {
+            /* create exec+connect services */
+            /* FIXME: this is just a crude workaround */
+            /*        is it better to kill the service? */
+            opt->option.retry=0;
         }
+        if(opt->ctx) {
+            s_log(LOG_DEBUG, "Sessions cached before flush: %ld",
+                SSL_CTX_sess_number(opt->ctx));
+            SSL_CTX_flush_sessions(opt->ctx,
+                (long)time(NULL)+opt->session_timeout+1);
+            s_log(LOG_DEBUG, "Sessions cached after flush: %ld",
+                SSL_CTX_sess_number(opt->ctx));
+        }
+        s_log(LOG_DEBUG, "Service [%s] closed", opt->servname);
+    }
 }
 
 /* open new ports, update fds */
@@ -268,6 +361,13 @@ int bind_ports(void) {
 
     s_poll_init(fds);
     s_poll_add(fds, signal_pipe[0], 1, 0);
+
+    /* allow clean unbind_ports() even though
+       bind_ports() was not fully performed */
+    for(opt=service_options.next; opt; opt=opt->next)
+        if(opt->option.accept)
+            opt->fd=-1;
+
     for(opt=service_options.next; opt; opt=opt->next) {
         if(opt->option.accept) {
             opt->fd=s_socket(opt->local_addr.sa.sa_family,
@@ -281,93 +381,52 @@ int bind_ports(void) {
             /* local socket can't be unnamed */
             local_address=s_ntop(&opt->local_addr, addr_len(&opt->local_addr));
             if(bind(opt->fd, &opt->local_addr.sa, addr_len(&opt->local_addr))) {
-                s_log(LOG_ERR, "Error binding %s to %s",
+                s_log(LOG_ERR, "Error binding service [%s] to %s",
                     opt->servname, local_address);
                 sockerror("bind");
                 closesocket(opt->fd);
                 str_free(local_address);
                 return 1;
             }
-            s_log(LOG_DEBUG, "Service %s bound to %s",
-                opt->servname, local_address);
-            str_free(local_address);
             if(listen(opt->fd, SOMAXCONN)) {
                 sockerror("listen");
                 closesocket(opt->fd);
+                str_free(local_address);
                 return 1;
             }
             s_poll_add(fds, opt->fd, 1, 0);
-            s_log(LOG_DEBUG, "Service %s opened FD=%d",
-                opt->servname, opt->fd);
+            s_log(LOG_DEBUG, "Service [%s] (FD=%d) bound to %s",
+                opt->servname, opt->fd, local_address);
+            str_free(local_address);
         } else if(opt->option.program && opt->option.remote) {
             /* create exec+connect services */
-            enter_critical_section(CRIT_CLIENTS);
-            ++num_clients;
-            leave_critical_section(CRIT_CLIENTS);
-            create_client(-1, -1, alloc_client_session(opt, -1, -1), client);
+            /* FIXME: needs to be delayed on reload with opt->option.retry set */
+            create_client(-1, -1,
+                alloc_client_session(opt, -1, -1), client_thread);
         }
     }
     return 0; /* OK */
 }
 
-static void get_limits(void) {
-    /* start with current ulimit */
-#if defined(HAVE_SYSCONF)
-    errno=0;
-    max_fds=sysconf(_SC_OPEN_MAX);
-    if(errno)
-        ioerror("sysconf");
-    if(max_fds<0)
-        max_fds=0; /* unlimited */
-#elif defined(HAVE_GETRLIMIT)
-    struct rlimit rlim;
-
-    if(getrlimit(RLIMIT_NOFILE, &rlim)<0) {
-        ioerror("getrlimit");
-        max_fds=0; /* unlimited */
-    } else
-        max_fds=rlim.rlim_cur!=RLIM_INFINITY ? rlim.rlim_cur : 0;
-#else
-    max_fds=0; /* unlimited */
-#endif /* HAVE_SYSCONF || HAVE_GETRLIMIT */
-
-#if !defined(USE_WIN32) && !defined(USE_POLL) && !defined(__INNOTEK_LIBC__)
-    /* apply FD_SETSIZE if select() is used on Unix */
-    if(!max_fds || max_fds>FD_SETSIZE)
-        max_fds=FD_SETSIZE; /* start with select() limit */
-#endif /* select() on Unix */
-
-    /* stunnel needs at least 16 file desriptors */
-    if(max_fds && max_fds<16)
-        max_fds=16;
-
-    if(max_fds) {
-        max_clients=max_fds>=256 ? max_fds*125/256 : (max_fds-6)/2;
-        s_log(LOG_DEBUG, "Clients allowed=%d", max_clients);
-    } else {
-        max_clients=0;
-        s_log(LOG_DEBUG, "No limit detected for the number of clients");
-    }
-}
-
 #ifdef HAVE_CHROOT
-static void change_root(void) {
-    if(global_options.chroot_dir) {
-        if(chroot(global_options.chroot_dir)) {
-            sockerror("chroot");
-            die(1);
-        }
-        if(chdir("/")) {
-            sockerror("chdir");
-            die(1);
-        }
+static int change_root(void) {
+    if(!global_options.chroot_dir)
+        return 0;
+    if(chroot(global_options.chroot_dir)) {
+        sockerror("chroot");
+        return 1;
     }
+    if(chdir("/")) {
+        sockerror("chdir");
+        return 1;
+    }
+    return 0;
 }
 #endif /* HAVE_CHROOT */
 
 #if !defined(USE_WIN32) && !defined(__vms) && !defined(USE_OS2)
 
-void drop_privileges(int critical) {
+int drop_privileges(int critical) {
 #ifdef HAVE_SETGROUPS
     gid_t gr_list[1];
 #endif
@@ -376,64 +435,68 @@ void drop_privileges(int critical) {
     if(global_options.gid) {
         if(setgid(global_options.gid) && critical) {
             sockerror("setgid");
-            die(1);
+            return 1;
         }
 #ifdef HAVE_SETGROUPS
         gr_list[0]=global_options.gid;
         if(setgroups(1, gr_list) && critical) {
             sockerror("setgroups");
-            die(1);
+            return 1;
         }
 #endif
     }
     if(global_options.uid) {
         if(setuid(global_options.uid) && critical) {
             sockerror("setuid");
-            die(1);
+            return 1;
         }
     }
+    return 0;
 }
 
-static void daemonize(void) { /* go to background */
-    close(0);
-    close(1);
-    close(2);
+static int daemonize(int fd) { /* go to background */
+    if(global_options.option.foreground)
+        return 0;
+    dup2(fd, 0);
+    dup2(fd, 1);
+    dup2(fd, 2);
 #if defined(HAVE_DAEMON) && !defined(__BEOS__)
     /* set noclose option when calling daemon() function,
      * so it does not require /dev/null device in the chrooted directory */
     if(daemon(0, 1)==-1) {
         ioerror("daemon");
-        die(1);
+        return 1;
     }
 #else
     chdir("/");
     switch(fork()) {
     case -1:    /* fork failed */
         ioerror("fork");
-        die(1);
+        return 1;
     case 0:     /* child */
         break;
     default:    /* parent */
-        die(0);
+        exit(0);
     }
 #endif
 #ifdef HAVE_SETSID
     setsid(); /* ignore the error */
 #endif
+    return 0;
 }
 
-static void create_pid(void) {
+static int create_pid(void) {
     int pf;
     char *pid;
 
     if(!global_options.pidfile) {
         s_log(LOG_DEBUG, "No pid file being created");
-        return;
+        return 0;
     }
     if(global_options.pidfile[0]!='/') {
         /* to prevent creating pid file relative to '/' after daemonize() */
         s_log(LOG_ERR, "Pid file (%s) must be full path name", global_options.pidfile);
-        die(1);
+        return 1;
     }
     global_options.dpid=(unsigned long)getpid();
 
@@ -443,7 +506,7 @@ static void create_pid(void) {
     if(pf==-1) {
         s_log(LOG_ERR, "Cannot create pid file %s", global_options.pidfile);
         ioerror("create");
-        die(1);
+        return 1;
     }
     pid=str_printf("%lu\n", global_options.dpid);
     write(pf, pid, strlen(pid));
@@ -451,6 +514,7 @@ static void create_pid(void) {
     close(pf);
     s_log(LOG_DEBUG, "Created pid file %s", global_options.pidfile);
     atexit(delete_pid);
+    return 0;
 }
 
 static void delete_pid(void) {
@@ -465,13 +529,12 @@ static void delete_pid(void) {
 
 /**************************************** signal pipe handling */
 
-static void signal_pipe_init(void) {
+static int signal_pipe_init(void) {
 #ifdef USE_WIN32
     if(make_sockets(signal_pipe))
-        die(1);
-#else
-#if defined(__INNOTEK_LIBC__)
-    /* Innotek port of GCC can not use select on a pipe
+        return 1;
+#elif defined(__INNOTEK_LIBC__)
+    /* Innotek port of GCC can not use select on a pipe:
      * use local socket instead */
     struct sockaddr_un un;
     fd_set set_pipe;
@@ -497,32 +560,20 @@ static void signal_pipe_init(void) {
         closesocket(pipe_in);
     } else {
         sockerror("select");
-        die(1);
+        return 1;
     }
-#else /* __INNOTEK_LIBC__ */
+#else /* Unix */
     if(s_pipe(signal_pipe, 1, "signal_pipe"))
-        die(1);
-#endif /* __INNOTEK_LIBC__ */
-
-    signal(SIGCHLD, signal_handler); /* a child has died */
-    signal(SIGHUP, signal_handler); /* configuration reload */
-    signal(SIGUSR1, signal_handler); /* log reopen */
-    signal(SIGPIPE, SIG_IGN); /* ignore "broken pipe" */
-    if(signal(SIGTERM, SIG_IGN)!=SIG_IGN)
-        signal(SIGTERM, signal_handler); /* fatal */
-    if(signal(SIGQUIT, SIG_IGN)!=SIG_IGN)
-        signal(SIGQUIT, signal_handler); /* fatal */
-    if(signal(SIGINT, SIG_IGN)!=SIG_IGN)
-        signal(SIGINT, signal_handler); /* fatal */
-    /* signal(SIGSEGV, signal_handler); */
+        return 1;
 #endif /* USE_WIN32 */
+    return 0;
 }
 
 void signal_post(int sig) {
     writesocket(signal_pipe[1], (char *)&sig, sizeof sig);
 }
 
-static void signal_pipe_empty(void) {
+static int signal_pipe_dispatch(void) {
     int sig, err;
 
     s_log(LOG_DEBUG, "Dispatching signals from the signal pipe");
@@ -552,9 +603,6 @@ static void signal_pipe_empty(void) {
                     /* FIXME: handle the error */
                 }
             }
-#ifdef USE_WIN32
-            win_newconfig();
-#endif
             break;
         case SIGNAL_REOPEN_LOG:
             s_log(LOG_DEBUG, "Processing SIGNAL_REOPEN_LOG");
@@ -565,13 +613,14 @@ static void signal_pipe_empty(void) {
         case SIGNAL_TERMINATE:
             s_log(LOG_DEBUG, "Processing SIGNAL_TERMINATE");
             s_log(LOG_NOTICE, "Terminated");
-            die(2);
+            return 2;
         default:
             s_log(LOG_ERR, "Received signal %d; terminating", sig);
-            die(1);
+            return 1;
         }
     }
     s_log(LOG_DEBUG, "Signal pipe is empty");
+    return 0;
 }
 
 #ifdef USE_FORK
@@ -580,23 +629,21 @@ static void client_status(void) { /* dead children detected */
 
 #ifdef HAVE_WAIT_FOR_PID
     while((pid=wait_for_pid(-1, &status, WNOHANG))>0) {
-        --num_clients; /* one client less */
 #else
     if((pid=wait(&status))>0) {
-        --num_clients; /* one client less */
 #endif
 #ifdef WIFSIGNALED
         if(WIFSIGNALED(status)) {
-            s_log(LOG_DEBUG, "Process %d terminated on signal %d (%d left)",
-                pid, WTERMSIG(status), num_clients);
+            s_log(LOG_DEBUG, "Process %d terminated on signal %d",
+                pid, WTERMSIG(status));
         } else {
-            s_log(LOG_DEBUG, "Process %d finished with code %d (%d left)",
-                pid, WEXITSTATUS(status), num_clients);
+            s_log(LOG_DEBUG, "Process %d finished with code %d",
+                pid, WEXITSTATUS(status));
         }
     }
 #else
-        s_log(LOG_DEBUG, "Process %d finished with code %d (%d left)",
-            pid, status, num_clients);
+        s_log(LOG_DEBUG, "Process %d finished with code %d",
+            pid, status);
     }
 #endif
 }
@@ -638,152 +685,7 @@ static void signal_handler(int sig) {
 
 #endif /* !defined(USE_WIN32) && !defined(USE_OS2) */
 
-/**************************************** file descriptor validation */
-
-#if defined USE_NEW_LINUX_API && defined HAVE_SOCKET4
-#define USE_NEW_LINUX_API 1
-#endif
-
-int s_socket(int domain, int type, int protocol, int nonblock, char *msg) {
-#ifdef USE_NEW_LINUX_API
-    if(nonblock)
-        type|=SOCK_NONBLOCK|SOCK_CLOEXEC;
-    else
-        type|=SOCK_CLOEXEC;
-#endif
-    return setup_fd(socket(domain, type, protocol), nonblock, msg);
-}
-
-int s_accept(int sockfd, struct sockaddr *addr, socklen_t *addrlen,
-        int nonblock, char *msg) {
-    int fd;
-
-#ifdef USE_NEW_LINUX_API
-    if(nonblock)
-        fd=accept4(sockfd, addr, addrlen, SOCK_NONBLOCK|SOCK_CLOEXEC);
-    else
-        fd=accept4(sockfd, addr, addrlen, SOCK_CLOEXEC);
-#else
-    fd=accept(sockfd, addr, addrlen);
-#endif
-    return setup_fd(fd, nonblock, msg);
-}
-
-#ifndef USE_WIN32
-
-int s_socketpair(int domain, int type, int protocol, int sv[2],
-        int nonblock, char *msg) {
-#ifdef USE_NEW_LINUX_API
-    if(nonblock)
-        type|=SOCK_NONBLOCK|SOCK_CLOEXEC;
-    else
-        type|=SOCK_CLOEXEC;
-#endif
-    if(socketpair(domain, type, protocol, sv)<0) {
-        ioerror(msg);
-        return -1;
-    }
-    if(setup_fd(sv[0], nonblock, msg)<0) {
-        closesocket(sv[1]);
-        return -1;
-    }
-    if(setup_fd(sv[1], nonblock, msg)<0) {
-        closesocket(sv[0]);
-        return -1;
-    }
-    return 0;
-}
-
-int s_pipe(int pipefd[2], int nonblock, char *msg) {
-    int retval;
-
-#ifdef USE_NEW_LINUX_API
-    if(nonblock)
-        retval=pipe2(pipefd, O_NONBLOCK|O_CLOEXEC);
-    else
-        retval=pipe2(pipefd, O_CLOEXEC);
-#else
-    retval=pipe(pipefd);
-#endif
-    if(retval<0) {
-        ioerror(msg);
-        return -1;
-    }
-    if(setup_fd(pipefd[0], nonblock, msg)<0) {
-        close(pipefd[1]);
-        return -1;
-    }
-    if(setup_fd(pipefd[1], nonblock, msg)<0) {
-        close(pipefd[0]);
-        return -1;
-    }
-    return 0;
-}
-
-#endif /* USE_WIN32 */
-
-/* try to use non-POSIX O_NDELAY on obsolete BSD systems */
-#if !defined O_NONBLOCK && defined O_NDELAY
-#define O_NONBLOCK O_NDELAY
-#endif
-
-static int setup_fd(int fd, int nonblock, char *msg) {
-#ifdef FD_CLOEXEC
-    int err;
-#endif /* FD_CLOEXEC */
-
-    if(fd<0) {
-        sockerror(msg);
-        return -1;
-    }
-    if(max_fds && fd>=max_fds) {
-        s_log(LOG_ERR,
-            "%s: FD=%d out of range (max %d)", msg, fd, max_fds);
-        closesocket(fd);
-        return -1;
-    }
-#ifndef USE_NEW_LINUX_API
-    set_nonblock(fd, nonblock);
-#ifdef FD_CLOEXEC
-    do {
-        err=fcntl(fd, F_SETFD, FD_CLOEXEC);
-    } while(err<0 && get_last_socket_error()==S_EINTR);
-    if(err<0)
-        sockerror("fcntl SETFD"); /* non-critical */
-#endif /* FD_CLOEXEC */
-#endif /* USE_NEW_LINUX_API */
-    s_log(LOG_DEBUG, "%s: FD=%d allocated (%sblocking mode)",
-        msg, fd, nonblock?"non-":"");
-    return fd;
-}
-
-void set_nonblock(int fd, unsigned long nonblock) {
-#if defined F_GETFL && defined F_SETFL && defined O_NONBLOCK && !defined __INNOTEK_LIBC__
-    int err, flags;
-
-    do {
-        flags=fcntl(fd, F_GETFL, 0);
-    } while(flags<0 && get_last_socket_error()==S_EINTR);
-    if(flags<0) {
-        sockerror("fcntl GETFL"); /* non-critical */
-        return;
-    }
-    if(nonblock)
-        flags|=O_NONBLOCK;
-    else
-        flags&=~O_NONBLOCK;
-    do {
-        err=fcntl(fd, F_SETFL, flags);
-    } while(err<0 && get_last_socket_error()==S_EINTR);
-    if(err<0)
-        sockerror("fcntl SETFL"); /* non-critical */
-#else /* use fcntl() */
-    if(ioctlsocket(fd, FIONBIO, &nonblock)<0)
-        sockerror("ioctlsocket"); /* non-critical */
-#endif /* use fcntl() */
-}
-
-/**************************************** log messages to identify  build */
+/**************************************** log build details */
 
 void stunnel_info(int level) {
     s_log(level, "stunnel " STUNNEL_VERSION " on " HOST " platform");
@@ -795,6 +697,7 @@ void stunnel_info(int level) {
         s_log(level, "Update OpenSSL shared libraries or rebuild stunnel");
     }
     s_log(level,
+
         "Threading:"
 #ifdef USE_UCONTEXT
         "UCONTEXT"
@@ -809,30 +712,37 @@ void stunnel_info(int level) {
         "FORK"
 #endif
 
-        " SSL:"
-#ifdef HAVE_OSSL_ENGINE_H
-        "ENGINE"
-#else /* defined(HAVE_OSSL_ENGINE_H) */
-        "NOENGINE"
-#endif /* defined(HAVE_OSSL_ENGINE_H) */
-#ifdef USE_FIPS
-        ",FIPS"
-#endif /* USE_FIPS */
-
-        " Auth:"
-#ifdef USE_LIBWRAP
-        "LIBWRAP"
-#else
-        "none"
-#endif
-
         " Sockets:"
 #ifdef USE_POLL
         "POLL"
 #else /* defined(USE_POLL) */
         "SELECT"
 #endif /* defined(USE_POLL) */
-        ",IPv%c",
+        ",IPv%c"
+
+#if defined HAVE_OSSL_ENGINE_H || defined HAVE_OSSL_OCSP_H || defined USE_FIPS
+        " SSL:"
+#define ITEM_SEPARATOR ""
+#ifdef HAVE_OSSL_ENGINE_H
+        "ENGINE"
+#undef ITEM_SEPARATOR
+#define ITEM_SEPARATOR ","
+#endif /* defined(HAVE_OSSL_ENGINE_H) */
+#ifdef HAVE_OSSL_OCSP_H
+        ITEM_SEPARATOR "OCSP"
+#undef ITEM_SEPARATOR
+#define ITEM_SEPARATOR ","
+#endif /* HAVE_OSSL_OCSP_H */
+#ifdef USE_FIPS
+        ITEM_SEPARATOR "FIPS"
+#endif /* USE_FIPS */
+#endif /* an SSL optional feature enabled */
+
+#ifdef USE_LIBWRAP
+        " Auth:LIBWRAP"
+#endif
+
+        , /* supported IP version parameter */
 #if defined(USE_WIN32) && !defined(_WIN32_WCE)
         s_getaddrinfo ? '6' : '4'
 #else /* defined(USE_WIN32) */
@@ -845,18 +755,4 @@ void stunnel_info(int level) {
         );
 }
 
-/**************************************** fatal error */
-
-void die(int status) { /* some cleanup and exit */
-    unbind_ports();
-    s_poll_free(fds);
-    str_stats();
-    log_flush(LOG_MODE_ERROR);
-#ifdef USE_WIN32
-    win_exit(status);
-#else
-    exit(status);
-#endif
-}
-
 /* end of stunnel.c */
